from typing import List, Optional

import json
import logging
import os
import time
import uuid

import redis
from attr import asdict

from parea.cache.cache import Cache
from parea.schemas.models import CacheRequest, TraceLog

logger = logging.getLogger()


def is_uuid(value: str) -> bool:
    try:
        uuid.UUID(value)
    except ValueError:
        return False
    return True


class RedisCache(Cache):
    """A Redis-based cache for caching LLM responses."""

    def __init__(
        self,
        key_logs: str = os.getenv("_parea_redis_logs_key", f"trace-logs-{time.time()}"),
        host: str = os.getenv("REDIS_HOST", "localhost"),
        port: int = int(os.getenv("REDIS_PORT", 6379)),
        password: str = os.getenv("REDIS_PASSWORT", None),
        ttl=3600 * 6,
    ):
        """
        Initialize the cache.

        Args:
            key_logs (str): The Redis key for the logs.
            host (str): The Redis host.
            port (int): The Redis port.
            password (str): The Redis password.
            ttl (int): The default TTL for cache entries, in seconds.
        """
        self.r = redis.Redis(
            host=host,
            port=port,
            password=password,
        )
        self.ttl = ttl
        self.key_logs = key_logs

    def get(self, key: CacheRequest) -> Optional[TraceLog]:
        try:
            result = self.r.get(json.dumps(asdict(key)))
            if result is not None:
                return TraceLog(**json.loads(result))
        except redis.RedisError as e:
            logger.error(f"Error getting key {key} from cache: {e}")
        return None

    async def aget(self, key: CacheRequest) -> Optional[TraceLog]:
        return self.get(key)

    def set(self, key: CacheRequest, value: TraceLog):
        try:
            self.r.set(json.dumps(asdict(key)), json.dumps(asdict(value)), ex=self.ttl)
        except redis.RedisError as e:
            logger.error(f"Error setting key {key} in cache: {e}")

    async def aset(self, key: CacheRequest, value: TraceLog):
        self.set(key, value)

    def invalidate(self, key: CacheRequest):
        """
        Invalidate a key in the cache.

        Args:
            key (CacheRequest): The cache key.
        """
        try:
            self.r.delete(json.dumps(asdict(key)))
        except redis.RedisError as e:
            logger.error(f"Error invalidating key {key} from cache: {e}")

    async def ainvalidate(self, key: CacheRequest):
        self.invalidate(key)

    def log(self, value: TraceLog):
        try:
            prev_logs = self.r.hget(self.key_logs, value.trace_id)
            log_dict = asdict(value)
            if prev_logs:
                log_dict = {**json.loads(prev_logs), **log_dict}
            self.r.hset(self.key_logs, value.trace_id, json.dumps(log_dict))
        except redis.RedisError as e:
            logger.error(f"Error adding to list in cache: {e}")

    def read_logs(self) -> List[TraceLog]:
        try:
            trace_logs_raw = self.r.hgetall(self.key_logs)
            trace_logs = []
            for trace_log_raw in trace_logs_raw.values():
                trace_logs.append(TraceLog(**json.loads(trace_log_raw)))
            return trace_logs
        except redis.RedisError as e:
            logger.error(f"Error reading list from cache: {e}")
        return []



from attrs import asdict, define, field

from parea.api_client import HTTPClient
from parea.cache.redis import RedisCache
from parea.schemas.models import TraceLog

LOG_ENDPOINT = "/trace_log"


@define
class PareaLogger:
    _client: HTTPClient = field(init=False, default=None)
    _redis_lru_cache: RedisCache = field(init=False, default=None)

    def set_client(self, client: HTTPClient) -> None:
        self._client = client

    def set_redis_lru_cache(self, cache: RedisCache) -> None:
        self._redis_lru_cache = cache

    def record_log(self, data: TraceLog) -> None:
        self._client.request(
            "POST",
            LOG_ENDPOINT,
            data=asdict(data),
        )

    async def arecord_log(self, data: TraceLog) -> None:
        await self._client.request_async(
            "POST",
            LOG_ENDPOINT,
            data=asdict(data),
        )

    def write_log(self, data: TraceLog) -> None:
        self._redis_lru_cache.log(data)

    def default_log(self, data: TraceLog) -> None:
        if self._redis_lru_cache:
            self.write_log(data)
        if self._client:
            self.record_log(data)


parea_logger = PareaLogger()



from typing import Any, AsyncIterator, Callable, Dict, Iterator, Optional, Sequence, Union

import json
from collections import defaultdict

import openai
from openai.openai_object import OpenAIObject
from openai.util import convert_to_openai_object

from ..cache.cache import Cache
from ..schemas.models import CacheRequest, LLMInputs, ModelParams, TraceLog
from ..utils.trace_utils import trace_data
from .wrapper import Wrapper

MODEL_COST_MAPPING: Dict[str, float] = {
    "gpt-4": 0.03,
    "gpt-4-0314": 0.03,
    "gpt-4-0613": 0.03,
    "gpt-4-completion": 0.06,
    "gpt-4-0314-completion": 0.06,
    "gpt-4-0613-completion": 0.06,
    "gpt-4-32k": 0.06,
    "gpt-4-32k-0314": 0.06,
    "gpt-4-32k-0613": 0.06,
    "gpt-4-32k-completion": 0.12,
    "gpt-4-32k-0314-completion": 0.12,
    "gpt-4-32k-0613-completion": 0.12,
    "gpt-3.5-turbo": 0.0015,
    "gpt-3.5-turbo-0301": 0.0015,
    "gpt-3.5-turbo-0613": 0.0015,
    "gpt-3.5-turbo-16k": 0.003,
    "gpt-3.5-turbo-16k-0613": 0.003,
    "gpt-3.5-turbo-completion": 0.002,
    "gpt-3.5-turbo-0301-completion": 0.002,
    "gpt-3.5-turbo-0613-completion": 0.004,
    "gpt-3.5-turbo-16k-completion": 0.004,
    "gpt-3.5-turbo-16k-0613-completion": 0.004,
    "text-ada-001": 0.0004,
    "ada": 0.0004,
    "text-babbage-001": 0.0005,
    "babbage": 0.0005,
    "text-curie-001": 0.002,
    "curie": 0.002,
    "text-davinci-003": 0.02,
    "text-davinci-002": 0.02,
    "code-davinci-002": 0.02,
}


class OpenAIWrapper:
    original_methods = {"ChatCompletion.create": openai.ChatCompletion.create, "ChatCompletion.acreate": openai.ChatCompletion.acreate}

    def init(self, log: Callable, cache: Cache = None):
        Wrapper(
            resolver=self.resolver,
            gen_resolver=self.gen_resolver,
            agen_resolver=self.agen_resolver,
            log=log,
            module=openai,
            func_names=list(self.original_methods.keys()),
            cache=cache,
            convert_kwargs_to_cache_request=self.convert_kwargs_to_cache_request,
            convert_cache_to_response=self.convert_cache_to_response,
            aconvert_cache_to_response=self.aconvert_cache_to_response,
        )

    @staticmethod
    def resolver(trace_id: str, _args: Sequence[Any], kwargs: Dict[str, Any], response: Optional[Any]) -> Optional[Any]:
        if response:
            usage = response["usage"]
            output = OpenAIWrapper._get_output(response)
        else:
            output = None
            usage = {}

        llm_configuration = OpenAIWrapper._kwargs_to_llm_configuration(kwargs)
        model = llm_configuration.model

        model_rate = OpenAIWrapper.get_model_cost(model)
        model_completion_rate = OpenAIWrapper.get_model_cost(model, is_completion=True)
        completion_cost = model_completion_rate * (usage.get("completion_tokens", 0) / 1000)
        prompt_cost = model_rate * (usage.get("prompt_tokens", 0) / 1000)
        total_cost = sum([prompt_cost, completion_cost])

        trace_data.get()[trace_id].configuration = llm_configuration
        trace_data.get()[trace_id].input_tokens = usage.get("prompt_tokens", 0)
        trace_data.get()[trace_id].output_tokens = usage.get("completion_tokens", 0)
        trace_data.get()[trace_id].total_tokens = usage.get("total_tokens", 0)
        trace_data.get()[trace_id].cost = total_cost
        trace_data.get()[trace_id].output = output
        return response

    @staticmethod
    def gen_resolver(trace_id: str, _args: Sequence[Any], kwargs: Dict[str, Any], response: Iterator[Any], final_log) -> Iterator[Any]:
        llm_configuration = OpenAIWrapper._kwargs_to_llm_configuration(kwargs)
        trace_data.get()[trace_id].configuration = llm_configuration

        message = defaultdict(str)
        for chunk in response:
            update_dict = chunk.choices[0].delta._previous
            for key, val in update_dict.items():
                if val is not None:
                    message[key] += str(val)
            yield chunk

        trace_data.get()[trace_id].output = OpenAIWrapper._get_output(message)

        final_log()

    @staticmethod
    async def agen_resolver(trace_id: str, _args: Sequence[Any], kwargs: Dict[str, Any], response: AsyncIterator[Any], final_log) -> AsyncIterator[Any]:
        llm_configuration = OpenAIWrapper._kwargs_to_llm_configuration(kwargs)
        trace_data.get()[trace_id].configuration = llm_configuration

        message = defaultdict(str)
        async for chunk in response:
            update_dict = chunk.choices[0].delta._previous
            for key, val in update_dict.items():
                if val is not None:
                    message[key] += str(val)
            yield chunk

        trace_data.get()[trace_id].output = OpenAIWrapper._get_output(message)

        final_log()

    @staticmethod
    def _kwargs_to_llm_configuration(kwargs):
        return LLMInputs(
            model=kwargs.get("model", None),
            provider="openai",
            messages=kwargs.get("messages", None),
            functions=kwargs.get("functions", None),
            function_call=kwargs.get("function_call", None),
            model_params=ModelParams(
                temp=kwargs.get("temperature", 1.0),
                max_length=kwargs.get("max_tokens", None),
                top_p=kwargs.get("top_p", 1.0),
                frequency_penalty=kwargs.get("frequency_penalty", 0.0),
                presence_penalty=kwargs.get("presence_penalty", 0.0),
            ),
        )

    @staticmethod
    def _get_output(result: Any) -> str:
        if not isinstance(result, OpenAIObject) and isinstance(result, dict):
            result = convert_to_openai_object(
                {
                    "choices": [
                        {
                            "index": 0,
                            "message": result,
                        }
                    ]
                }
            )
        response_message = result.choices[0].message
        if response_message.get("function_call", None):
            completion = OpenAIWrapper._format_function_call(response_message)
        else:
            completion = response_message.content.strip()
        return completion

    @staticmethod
    def _format_function_call(response_message) -> str:
        function_name = response_message["function_call"]["name"]
        if isinstance(response_message["function_call"]["arguments"], OpenAIObject):
            function_args = dict(response_message["function_call"]["arguments"])
        else:
            function_args = json.loads(response_message["function_call"]["arguments"])
        return json.dumps({"name": function_name, "arguments": function_args}, indent=4)

    @staticmethod
    def get_model_cost(model_name: str, is_completion: bool = False) -> float:
        model_name = model_name.lower()

        if model_name.startswith("gpt-4") and is_completion:
            model_name += "-completion"

        cost = MODEL_COST_MAPPING.get(model_name, None)
        if cost is None:
            msg = f"Unknown model: {model_name}. " f"Please provide a valid OpenAI model name. " f"Known models are: {', '.join(MODEL_COST_MAPPING.keys())}"
            raise ValueError(msg)

        return cost

    @staticmethod
    def convert_kwargs_to_cache_request(_args: Sequence[Any], kwargs: Dict[str, Any]) -> CacheRequest:
        return CacheRequest(
            configuration=OpenAIWrapper._kwargs_to_llm_configuration(kwargs),
        )

    @staticmethod
    def _convert_cache_to_response(_args: Sequence[Any], kwargs: Dict[str, Any], cache_response: TraceLog) -> OpenAIObject:
        content = cache_response.output
        message = {"role": "assistant"}
        try:
            function_call = json.loads(content)
            if isinstance(function_call, dict) and "name" in function_call and "arguments" in function_call and len(function_call) == 2:
                message["function_call"] = function_call
                message["content"] = None
            else:
                message["content"] = content
        except json.JSONDecodeError:
            message["content"] = content

        message_field = "delta" if kwargs.get("stream", False) else "message"

        return convert_to_openai_object(
            {
                "object": "chat.completion",
                "model": cache_response.configuration["model"],
                "choices": [
                    {
                        "index": 0,
                        message_field: message,
                    }
                ],
                "usage": {
                    "prompt_tokens": cache_response.input_tokens,
                    "completion_tokens": cache_response.output_tokens,
                    "total_tokens": cache_response.total_tokens,
                },
            }
        )

    @staticmethod
    def convert_cache_to_response(_args: Sequence[Any], kwargs: Dict[str, Any], cache_response: TraceLog) -> Union[OpenAIObject, Iterator[OpenAIObject]]:
        response = OpenAIWrapper._convert_cache_to_response(_args, kwargs, cache_response)
        if kwargs.get("stream", False):
            return iter([response])
        else:
            return response

    @staticmethod
    def aconvert_cache_to_response(_args: Sequence[Any], kwargs: Dict[str, Any], cache_response: TraceLog) -> Union[OpenAIObject, AsyncIterator[OpenAIObject]]:
        response = OpenAIWrapper._convert_cache_to_response(_args, kwargs, cache_response)
        if kwargs.get("stream", False):

            def aiterator(iterable):
                async def gen():
                    for item in iterable:
                        yield item

                return gen()

            return aiterator([response])
        else:
            return response



import argparse
import asyncio
import concurrent
import csv
import importlib
import inspect
import os
import sys
import time
from importlib import util

from attr import asdict, fields_dict
from tqdm import tqdm

from parea.cache.redis import RedisCache
from parea.schemas.models import TraceLog


def load_from_path(module_path, attr_name):
    # Ensure the directory of user-provided script is in the system path
    dir_name = os.path.dirname(module_path)
    if dir_name not in sys.path:
        sys.path.insert(0, dir_name)

    module_name = os.path.basename(module_path)
    # Add .py extension back in to allow import correctly
    module_path_with_ext = f"{module_path}.py"

    spec = importlib.util.spec_from_file_location(module_name, module_path_with_ext)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    if spec.name not in sys.modules:
        sys.modules[spec.name] = module

    fn = getattr(module, attr_name)
    return fn


def read_input_file(file_path) -> list[dict]:
    with open(file_path) as file:
        reader = csv.DictReader(file)
        inputs = list(reader)
    return inputs


def async_wrapper(fn, **kwargs):
    return asyncio.run(fn(**kwargs))


def run_benchmark(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("--func", help="Function to test e.g., path/to/my_code.py:argument_chain", type=str)
    parser.add_argument("--csv_path", help="Path to the input CSV file", type=str)
    parser.add_argument("--redis_host", help="Redis host", type=str, default=os.getenv("REDIS_HOST", "localhost"))
    parser.add_argument("--redis_port", help="Redis port", type=int, default=int(os.getenv("REDIS_PORT", 6379)))
    parser.add_argument("--redis_password", help="Redis password", type=str, default=None)
    parsed_args = parser.parse_args(args)

    fn = load_from_path(*parsed_args.func.rsplit(":", 1))

    data_inputs = read_input_file(parsed_args.csv_path)

    redis_logs_key = f"parea-trace-logs-{int(time.time())}"
    os.putenv("_parea_redis_logs_key", redis_logs_key)

    with concurrent.futures.ProcessPoolExecutor() as executor:
        if inspect.iscoroutinefunction(fn):
            futures = [executor.submit(async_wrapper, fn, **data_input) for data_input in data_inputs]
        else:
            futures = [executor.submit(fn, **data_input) for data_input in data_inputs]
        for f in tqdm(concurrent.futures.as_completed(futures), total=len(futures)):
            pass
        print(f"Done with {len(futures)} inputs")

        redis_cache = RedisCache(key_logs=redis_logs_key)

        trace_logs: list[TraceLog] = redis_cache.read_logs()

        # write to csv
        path_csv = f"trace_logs-{int(time.time())}.csv"
        with open(path_csv, "w", newline="") as file:
            # write header
            columns = fields_dict(TraceLog).keys()
            writer = csv.DictWriter(file, fieldnames=columns)
            writer.writeheader()
            # write rows
            for trace_log in trace_logs:
                writer.writerow(asdict(trace_log))

        print(f"Wrote CSV of results to: {path_csv}")



from typing import Optional

from abc import ABC

from parea.schemas.models import CacheRequest, TraceLog


class Cache(ABC):
    def get(self, key: CacheRequest) -> Optional[TraceLog]:
        """
        Get a normal response from the cache.

        Args:
            key (CacheRequest): The cache key.

        Returns:
            Optional[TraceLog]: The cached response, or None if the key was not found.

        # noqa: DAR202
        # noqa: DAR401
        """
        raise NotImplementedError

    async def aget(self, key: CacheRequest) -> Optional[TraceLog]:
        """
        Get a normal response from the cache.

        Args:
            key (CacheRequest): The cache key.

        Returns:
            Optional[TraceLog]: The cached response, or None if the key was not found.

        # noqa: DAR202
        # noqa: DAR401
        """
        raise NotImplementedError

    def set(self, key: CacheRequest, value: TraceLog):
        """
        Set a normal response in the cache.

        Args:
            key (CacheRequest): The cache key.
            value (TraceLog): The response to cache.

        # noqa: DAR401
        """
        raise NotImplementedError

    async def aset(self, key: CacheRequest, value: TraceLog):
        """
        Set a normal response in the cache.

        Args:
            key (CacheRequest): The cache key.
            value (TraceLog): The response to cache.

        # noqa: DAR401
        """
        raise NotImplementedError

    def invalidate(self, key: CacheRequest):
        """
        Invalidate a key in the cache.

        Args:
            key (CacheRequest): The cache key.

        # noqa: DAR401
        """
        raise NotImplementedError

    async def ainvalidate(self, key: CacheRequest):
        """
        Invalidate a key in the cache.

        Args:
            key (CacheRequest): The cache key.

        # noqa: DAR401
        """
        raise NotImplementedError



from typing import Callable

import asyncio
import os
import time

from attrs import asdict, define, field

from parea.api_client import HTTPClient
from parea.cache.cache import Cache
from parea.cache.redis import RedisCache
from parea.helpers import gen_trace_id
from parea.parea_logger import parea_logger
from parea.schemas.models import Completion, CompletionResponse, FeedbackRequest, UseDeployedPrompt, UseDeployedPromptResponse
from parea.utils.trace_utils import get_current_trace_id, logger_all_possible, logger_record_log, trace_data
from parea.wrapper import OpenAIWrapper

COMPLETION_ENDPOINT = "/completion"
DEPLOYED_PROMPT_ENDPOINT = "/deployed-prompt"
RECORD_FEEDBACK_ENDPOINT = "/feedback"


@define
class Parea:
    api_key: str = field(init=True, default="")
    _client: HTTPClient = field(init=False, default=HTTPClient())
    cache: Cache = field(init=True, default=None)

    def __attrs_post_init__(self):
        self._client.set_api_key(self.api_key)
        if self.api_key:
            parea_logger.set_client(self._client)
        if isinstance(self.cache, RedisCache):
            parea_logger.set_redis_lru_cache(self.cache)
        _init_parea_wrapper(logger_all_possible, self.cache)

    def completion(self, data: Completion) -> CompletionResponse:
        inference_id = gen_trace_id()
        data.inference_id = inference_id
        r = self._client.request(
            "POST",
            COMPLETION_ENDPOINT,
            data=asdict(data),
        )
        if parent_trace_id := get_current_trace_id():
            trace_data.get()[parent_trace_id].children.append(inference_id)
            logger_record_log(parent_trace_id)
        return CompletionResponse(**r.json())

    async def acompletion(self, data: Completion) -> CompletionResponse:
        inference_id = gen_trace_id()
        data.inference_id = inference_id
        r = await self._client.request_async(
            "POST",
            COMPLETION_ENDPOINT,
            data=asdict(data),
        )
        if parent_trace_id := get_current_trace_id():
            trace_data.get()[parent_trace_id].children.append(inference_id)
            logger_record_log(parent_trace_id)
        return CompletionResponse(**r.json())

    def get_prompt(self, data: UseDeployedPrompt) -> UseDeployedPromptResponse:
        r = self._client.request(
            "POST",
            DEPLOYED_PROMPT_ENDPOINT,
            data=asdict(data),
        )
        return UseDeployedPromptResponse(**r.json())

    async def aget_prompt(self, data: UseDeployedPrompt) -> UseDeployedPromptResponse:
        r = await self._client.request_async(
            "POST",
            DEPLOYED_PROMPT_ENDPOINT,
            data=asdict(data),
        )
        return UseDeployedPromptResponse(**r.json())

    def record_feedback(self, data: FeedbackRequest) -> None:
        time.sleep(2)  # give logs time to update
        self._client.request(
            "POST",
            RECORD_FEEDBACK_ENDPOINT,
            data=asdict(data),
        )

    async def arecord_feedback(self, data: FeedbackRequest) -> None:
        await asyncio.sleep(2)  # give logs time to update
        await self._client.request_async(
            "POST",
            RECORD_FEEDBACK_ENDPOINT,
            data=asdict(data),
        )


_initialized_parea_wrapper = False


def init(api_key: str = os.getenv("PAREA_API_KEY"), cache: Cache = None) -> None:
    Parea(api_key=api_key, cache=cache)


def _init_parea_wrapper(log: Callable = None, cache: Cache = None):
    global _initialized_parea_wrapper
    if _initialized_parea_wrapper:
        return
    OpenAIWrapper().init(log=log, cache=cache)
    _initialized_parea_wrapper = True
